package util.iterator.recursiv;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import util.coll.XArrayList;

/**
 * Beliebig tief schachtelbarer
 * rekursiver {@link Iterator},
 * bestehend aus
 * {@link RecursivOrTerminalIterator}en.
 * 
 * @param <E> Type-Parameter, Element-Type
 * @author Heiner K&uuml;cker
 */
public final class RecursivIterator<E>
implements RecursivOrTerminalIterator<E>
{
    /**
     * Iterator ber innere (untergeordnete) Iteratoren.
     */
	private final Iterator<? extends RecursivOrTerminalIterator<E>> innerIteratorIterator;

	/**
	 * aktueller untergeordneter Iterator
	 */
	private Iterator<E> currentSubIterator;

	/**
	 * Member mit remove-fhigem untergeordnetem Iterator.
	 * TODO Aufgabe noch mal klren, Unit-Test
	 */
	private Iterator<E> currentSubIteratorForRemove;

	/**
	 * Konstruktor.
	 * 
	 * @param iteratorIterator geschachtelter Iterator, Iterator ber Iteratoren
	 */
	public RecursivIterator(
			final Iterator<? extends RecursivOrTerminalIterator<E>> iteratorIterator )
	{
		this.innerIteratorIterator = iteratorIterator;
		
		// Sub-Iterator vor erstem Aufruf hasNext() bzw next() initialisieren
		if ( iteratorIterator.hasNext() )
		{
			this.currentSubIterator =
					iteratorIterator.next();
		}
	}

	/**
	 * Konstruktor.
	 * 
	 * @param iteratorVararg Vararg ber Sub-Iteratoren
	 */
	public RecursivIterator(
			final RecursivOrTerminalIterator<E>... iteratorVararg )
	{
		this.innerIteratorIterator =
				new XArrayList<RecursivOrTerminalIterator<E>>(
						iteratorVararg ).iterator();
		
		if ( this.innerIteratorIterator.hasNext() )
		{
			this.currentSubIterator =
					this.innerIteratorIterator.next();
		}
	}

	/**
	 * @see java.util.Iterator#hasNext()
	 */
	@Override
	public boolean hasNext()
	{
		return
				// das allererste Besorgen eines untergeordneten Iterators hat bereits nicht geklappt
				// oder es gab nach dem Verbrauchen eines Sub-Iterators keinen neuen
				this.currentSubIterator != null;
	}

	/**
	 * @see java.util.Iterator#next()
	 */
	@Override
	public E next()
	{
	    this.currentSubIteratorForRemove = null;

		if ( ! hasNext() )
		{
			throw new NoSuchElementException();
		}
		
		final E tmp = this.currentSubIterator.next();

		// Vermerken des aktuellen Iterators als remove-fhig
		this.currentSubIteratorForRemove = this.currentSubIterator;
		
		while ( this.currentSubIterator != null &&
				! this.currentSubIterator.hasNext() )
			// aktueller Iterator verbraucht
			// hier muss while rein, falls mehrere leere Iteratoren ankommen
		{
			if ( this.innerIteratorIterator.hasNext() )
			{
				this.currentSubIterator =
						this.innerIteratorIterator.next();
			}
			else
			{
				this.currentSubIterator = null;
			}
		}
		
		return tmp;
	}

	/**
	 * @see java.util.Iterator#remove()
	 */
	@Override
	public void remove()
	{
		if ( this.currentSubIteratorForRemove == null )
		{
			throw new IllegalStateException();
		}
		
		this.currentSubIteratorForRemove.remove();
	}

	/**
	 * 
	 * @param args
	 */
	public static void main(
			final String[] args )
	{
		final List<Integer> list0 =
				Arrays.asList(
						0 ,
						1 ,
						2 );
		
		final List<? extends RecursivOrTerminalIterator<Integer>> list1 =
				new ArrayList<RecursivOrTerminalIterator<Integer>>(
						Arrays.asList(
								new TerminalIterator<Integer>( list0.iterator() ) ,
								new TerminalIterator<Integer>( list0.iterator() ) ,
								new TerminalIterator<Integer>( list0.iterator() ) ) );
		
		final RecursivOrTerminalIterator<Integer> recursivIterator0 =
				new RecursivIterator<Integer>(
								list1.iterator() );
		
		while ( recursivIterator0.hasNext() )
		{
			System.out.println( recursivIterator0.next() );
		}
		
		System.out.println( "---------------------" );
		
		
		final List<? extends RecursivOrTerminalIterator<Integer>> list2 =
				new ArrayList<RecursivOrTerminalIterator<Integer>>(
						Arrays.asList(
								new RecursivIterator<Integer>(
										new ArrayList<RecursivOrTerminalIterator<Integer>>(
												Arrays.asList(
														new TerminalIterator<Integer>( list0.iterator() ) ,
														new TerminalIterator<Integer>( list0.iterator() ) ,
														new TerminalIterator<Integer>( list0.iterator() ) ) ).iterator() ) ,
								new RecursivIterator<Integer>(
										new ArrayList<RecursivOrTerminalIterator<Integer>>(
												Arrays.asList(
														new TerminalIterator<Integer>( list0.iterator() ) ,
														new TerminalIterator<Integer>( list0.iterator() ) ,
														new TerminalIterator<Integer>( list0.iterator() ) ) ).iterator() ) ,
								new RecursivIterator<Integer>(
										new ArrayList<RecursivOrTerminalIterator<Integer>>(
												Arrays.asList(
														new TerminalIterator<Integer>( list0.iterator() ) ,
														new TerminalIterator<Integer>( list0.iterator() ) ,
														new TerminalIterator<Integer>( list0.iterator() ) ) ).iterator() ) ) );

		final RecursivOrTerminalIterator<Integer> recursivIterator1 =
				new RecursivIterator<Integer>(
								list2.iterator() );

		while ( recursivIterator1.hasNext() )
		{
			System.out.println( recursivIterator1.next() );
		}
	}

}
